
%if 0

RxANSI: Combined ANSI and LESS driver. AMIS v3.6 compliant.
Public Domain by Henrik Haftmann and Christian Masloch

RxANSIT.ASM - Transient code of RxANSI; command line parser and setup

Read RxANSI.DOC for more information.

Assemble: NASM.EXE -fbin -oRXANSI.COM RXANSIT.ASM

Append "-d_NOI2F" or "-d_I2F" to assemble without interrupt 2Fh
hook or without option to disable the interrupt 2Fh hook.

(The following is a German notice)
Assemblieren: NASM.EXE -fbin -oRXANSI.COM RXANSIT.ASM

Die Zusatzparameter "-d_NOI2F" oder "-d_I2F" assemblieren eine
Version ohne Interrupt-2Fh-Umleitung oder ohne Option, die
Interrupt-2Fh-Umleitung auszuschalten. Der Zusatzparameter
"-d_RXANSI_DEU" assembliert die deutsche Version.

Hinweis:	Alle weiteren Kommentare in dieser Datei sind in Englisch.
		Die RxANSI-Dokumentation ist bisher ebenfalls Englisch, bei
		Bedarf kann ich aber auch eine deutsche Version erstellen.
		Kontaktdaten siehe am Ende der Datei RxANSI.DOC.

%endif

	%assign _186 0
	%assign _386 0

%include "CMMACROS.MAC"
%include "AMIS.MAC"

	verdef RXANSI,C, 1,10		; Version number of last compatible
	verdef RXANSI,P, 1,10		; Private version

	numdef NOI2F
	numdef I2F
	numdef RESUNINST
	numdef RESUNINSTSOC
	numdef DEU
	incdef NOI2F,RXANSI_NOI2F
	incdef I2F,RXANSI_I2F
	incdef RESUNINST,RXANSI_RESUNINST
	incdef RESUNINSTSOC,RXANSI_RESUNINSTSOC
	incdef DEU,RXANSI_DEU

	numdef UPXPAD			; UPX refuses to compress files <3062 bytes. Pad .COM to that size
	numdef RXANSI, 1, 1
	numdef RXBIO
	numdef RXANSI_RESUNINST		; include resident uninstaller
	numdef RXANSI_RESUNINSTSOC	; resident uninstaller: use alternative "soc" way
	numdef RXANSI_NOI2F		; Don't hook interrupt 2Fh
	numdef RXANSI_I2F		; Always hook interrupt 2Fh
	_incdef !(_RXANSI_NOI2F||_RXANSI_I2F),RXANSI_OPTI2F,1,0 ; Optionally hook interrupt 2Fh
	numdef RXANSI_DEU		; German messages (affects RxANSIT only)

%if _RXANSI_NOI2F && _RXANSI_I2F
 %error "NOI2F and I2F mustn't be true both"
%endif

%if _RXBIO
 %error "Include RXANSI.ASM for the internal version"
%endif


		; Bits of the option selector (hold in bp)
_holdenvironment	equ 0001h
_forcetoprelocation	equ 0002h
_installlow		equ 0004h
_installi2f		equ 0008h
_installnew		equ 0010h
_uninstall		equ 0020h
_lessstate		equ 0040h		; must match LESS bit in ctrl0
_ansistate		equ 0080h		; must match ANSI bit in ctrl0
_state			equ _ansistate|_lessstate
_lessset		equ _lessstate<<2
_ansiset		equ _ansistate<<2
_foundresident		equ 1000h
_incompatresident	equ 2000h

	[map all RXANSI.MAP]

	cpu 8086
	org 100h

	section TRANSIENT align=1	; Transient program part (command line parser, setup)

transient:
	cld
	mov dx, msg.name
	call disp_msg		; Start with message

	xor bp, bp		; initial option selector

	mov ax, cs
	add ax, 10h+transient_size_p
	mov ds, ax
	mov bx, rxansi_ctrl0	; ds:bx -> resident segment's ctrl0

findinstalled:
	mov ah, 0FFh		; start with multiplex number 0FFh
.loop:
	mov al, 00h		; AMIS installation check
	int 2Dh			; AMIS (or "DOS reserved" = iret if no AMIS present)
	cmp al, 0FFh
	je .try
.notfound:
	sub ah, 1		; search is backward (to find latest installed first), from 0FFh to 00h including
	jnc .loop		; try next if we didn't checked all yet -->

	jmp short scancl	;  If not found, start command line processing -->

.try:
	mov si, rxansi_amissig	; ds:si -> our AMIS name strings
	mov es, dx		; es:di -> name strings of AMIS multiplexer that just answered
	mov cx, 16		; Ignore description, only compare vendor and program name
	repe cmpsb
	jne .notfound		; No match, try next
	mov ds, dx		; set ds to our resident copy's segment
	mov al, 10h		; get compatible RxANSI version
	int 2Dh
	cmp al, 0FFh		; al = 0FFh if supported
	jne .unknown		; doesn't support this question -->
	cmp cx, RXANSI_VERC	; Compatible version?
	jne .unknown
	mov al, 20h		; Get RxANSI ctrl0
	int 2Dh
	cmp al, 00h		; al = size of returned ctrl data. at least one byte if supported
	jne .known		; if size ok, known -->
.unknown:	
	or bp, _incompatresident; indicate resident copy is incompatible
.known:
	or bp, _foundresident	; indicate resident copy found

scancl:			; bp = selected options
			; ah = AMIS multiplex number of resident copy if bp&_foundresident (else 0)
			; ds:bx-> ctrl0 of either resident copy or stored handler

	mov si, 81h		; PSP command line
	push cs
	pop es			; es changed while searching for resident
.loop:	cs lodsb		; ds is resident segment, so use cs
	cmp al, 9		; tab ?
	je .loop
	cmp al, 32		; blank ?
	je .loop
	jb .end			; zero, cr, lf ? -->
	cmp al, '-'
	je .switch
	cmp al, '/'
	jne .invalid		; unknown character -->
			; parse switch
.switch:
	cs lodsb		; get next character
	cmp al, 'a'
	jb .notlowercase
	cmp al, 'z'
	ja .notlowercase
	and al, ~ 20h		; Uppercase if required
.notlowercase:
			; check against known switches
	push si
	push ax
	mov si, parameters
	xchg al, ah
.loopswitch:
	cs lodsb
	xchg ax, cx
	cs lodsb
	xchg ax, cx		; cx = option selector bits
	cmp al, 0FFh
	je .invalid		; not found!
	cmp al, ah
	jne .loopswitch		; try next -->
	pop ax			; %$&" preserve ah (multiplex number) !
	pop si
	xor ch, ch

	test cl, _state		; ANSI/LESS state ?
	jz .normalbp		; no, normal option --> (sets bit in bp)

		; cx = bit to set or clear in bp
		; cx<<2 = bit to set in bp to show the cx bit was set
		; ds:si-> '+' to set, '-' to clear
	cs lodsb
	cmp	al,'-'		; OFF ?
	jne	.notminus
	call	flagset
	not	cx		; reverse bits
	and	bp,cx		; do bp&!cx
	jmp short .loop
.notminus:
	cmp	al,'+'		; ON ?
	jne	.invalid
	call	flagset
				; then fall through to do bp|cx
.normalbp:
	or bp, cx
	jmp short .loop		; ds:si-> character after switch character. just continue

.invalid:			; display help if invalid
	mov dx, msg.help
	jmp short abort_msg

.end:
	test bp, _uninstall
	jz .dontuninstall
	test bp, _forcetoprelocation|_holdenvironment| \
		_installlow|_installi2f|_installnew| \
		_ansiset|_lessset
	jnz .invalid		; all invalid if /u selected
	jmp short uninstall

.dontuninstall:
	test bp, _foundresident	; no resident found ?
	jz .install		; yes, install -->
	test bp, _installnew	; forced to install new ?
	jz .dontinstall		; no -->

.install:
	jmp install

.dontinstall:
	test bp, _forcetoprelocation|_holdenvironment| \
		_installlow|_installi2f
	jz .changeflags
	mov dx, msg.already	; can't install new, already resident
	jmp short abort_msg

.changeflags:
	test bp, _incompatresident
	jz .compatible
	mov dx, msg.incompatible ; can't change flags of incompatible version
	jmp short abort_msg

.compatible:
	call changeflags

		; Display status
	mov dx, msg.stateansi	; "state ANSI: "
	call disp_msg
	test byte [bx],80h
	call onoff		; "ON (+)" or "OFF (-)" depending on ZF
	mov dx, msg.stateless	; ", LESS: "
	call disp_msg
	test byte [bx],40h
	call onoff
	mov dx, msg.crlf
	call disp_msg
	mov ax, 4C00h
	int 21h

flagset:
	push cx
	rol cx, 1
	rol cx, 1
	or bp, cx
	pop cx
	retn

..@stack:			; above is overwritten by stack while installing
abort_msg:
	call disp_msg
	mov ax, 4CFFh
	int 21h

changeflags:
		; bp = option selector
		; ds:bx-> ctrl0
	mov ax, bp
	test ah, _ansiset>>8
	jz .noansi
	push ax
	and al, _ansistate
	and byte [bx], ~_ansistate
	or byte [bx], al
	pop ax
.noansi:
	test ah, _lessset>>8
	jz .noless
	and al, _lessstate
	and byte [bx], ~_lessstate
	or byte [bx], al
.noless:
	retn

uninstall:
	d4bp
	test bp, _foundresident
	jnz .found
	mov dx, msg.notinstalled	; nothing to uninstall there
.abort_msg:
	jmp short abort_msg

.found:
	test bp, _incompatresident
	jz .compat
	mov dx, msg.incompatible	; resident, but incompatible version
	jmp short abort_msg

.compat:
			; ds:bx-> ctrl0 of resident copy (unused yet)
			; ah = AMIS multiplex number of resident copy
	mov al, 02h
	mov dx, cs
	mov bx, .done			; dx:bx = return address if successful
	int 2Dh				; AMIS uninstall TSR
	cmp al, 0FFh			; 0FFh successful
	je .done			; TSR has already done anything -->
	cmp al, 00h			; 00h not implemented
	je .continue_noseg		; do it myself -->
	cmp al, 03h			; 03h safe, no resident uninstaller (still enabled). bx=segment
	je .continue_seg		; expected -->
	cmp al, 04h			; 04h safe, no resident uninstaller (now disabled). bx=segment
	je .continue_seg		; unexpected, but continue -->
		; (other values: 01h unsuccessful or internal, 02h can't uninstall yet,
		; but will do so when able, 05h not safe, 06h,07h device driver; all fail)
.fail:
	mov dx, msg.cantuninstall	; some error occured
	call disp_msg
	mov dl, '.'
	mov ah, 02h
	int 21h				; display dot to end message
	mov dx, msg.crlf
	jmp short .abort_msg

.continue_noseg:
	xor bx, bx			; = 0

.continue_seg:
	push bx

	mov al, 04h
	int 2Dh				; AMIS determine hooked interrupts
	cmp al, 04h
	jne .fail
	mov ds, dx			; ds:bx-> interrupt table
	push bx
	mov dx, unhookinterruptsim
	mov cx, .return1
	jmp short loopamisintr

.return1:
	jc .unhookerror
	pop bx				; restore ds:bx->

	mov dx, unhookinterrupt
	mov cx, .return2
	jmp short loopamisintr

.return2:
	jc .unhookerror

	push ds
	pop es				; es = segment of interrupt list
	pop ax				; stacked segment
	or ax, ax			; marker to use interrupt list segment ? (zero)
	jz .uselistsegment		;  yes -->
	mov es, ax			; else use segment returned by uninstall function

.uselistsegment:
	mov ah, 49h			; Free memory
	int 21h
				; Note:	General deinstallation code should not assume the TSR was just in
				;	that single memory block. If the deinstallation code doesn't know
				;	the TSR, it should search (and if it founds any, free) memory blocks
				;	that have the MCB owner either that's now in es (code segment of TSR)
				;	or the same MCB owner value that the TSR's code segments MCB had.
				;	(Only if the TSR's code segment was a valid memory block with MCB.)
	jc .fail			; if that causes an error, still report "fail"
					; (but interrupts are already unhooked now)

.done:
	mov dx, msg.removed
	call disp_msg			; "removed."
..@exit:
	mov ax, 4C00h
	int 21h

.unhookerror:
	mov dx, msg.cantuninstall
	call disp_msg
	mov dx, msg.interrupts		; first message

.unhookerrorloop:
	call disp_msg			; msg.interrupts or msg.comma
	call disp_al_hex
	pop ax				; get next value from stack
	cmp ax, byte -1			; last error marker?
	mov dx, msg.comma		; message if more interrupts to come
	jne .unhookerrorloop		; no, keep looping -->
	mov dx, msg.hooked
	jmp abort_msg

		; ds:bx -> AMIS interrupt list
		; dx = code called for each interrupt (with ds:si -> interrupt entry)
		; cx = code to return to
		;
		; OUT:	ss:sp-> 0FFFFh terminated list of failed interrupts on stack
		;	CY if ss:sp-> valid and any interrupts failed
		;
		; Called code (dx) has to preserve dx, bx, cx, al.
loopamisintr:
	
	mov ax, 0FFFFh
	push ax
.loop:
	mov al, [bx]
	mov si, [bx +1]

	call dx
	jnc .noerror			; no error -->
	xor ah, ah
	push ax				; else remember number of interrupt, but continue looping

.noerror:
	add bx, byte 3
	cmp al, 2Dh
	jne .loop			; do until 2Dh done -->

	pop ax
	cmp ax, byte -1			; if it's below (CY), there were errors
	jmp cx

		; INP:	al = interrupt number
		;	ds:si-> interrupt entry
		; OUT:	CY if unhooking failed
		;	NC if unhooking successful
unhookinterrupt:
			; unhookinterruptsim (below) only checks if it's possible to unhook this interrupt.
			; This function really unhooks the interrupt if possible.
			;
			; This is to cover the situation when some of the hooked interrupts can unhook,
			; but some can't. If the uninstaller would start to unhook the interrupts and then
			; catch the interrupt that can't be unhooked the user would end with a dead TSR
			; that's uninstalled halfway. Very bad.
			;
			; "Simulating" the unhooking first and checking if all Interrupts can unhook
			; normally will not return such a state.
	call unhookinterruptsim
	jc .ret				; bad. -->
	jz .easy
.hard:
				; "hard" case: unhookinterruptsim has however already done the work,
				; so the hard case is here indeed easier than the easy case.
	cmpsw				; compares the word of 10EBh to 10EBh. This sets NC, ZR
	movsw				; copies our stored pointer into the other's entry
	movsw				; copies our stored segment into the other's entry
	retn
.easy:
	push ds
	push dx
	lds dx, [ si + 2 ]		; get what we stored in the entry
	mov ah, 25h			; easy case - just reset to the value stored
	int 21h				; doesn't alter CF (leaves NC from unhookinterruptsim) or sets NC
	pop dx
	pop ds
.ret:
	retn

		; INP:	ds:si-> IISP interrupt entry
		;	al = interrupt number
		; OUT:	NC if no error (either hard or easy case)
		;	 ZR if easy case: ds:si-> our IISP interrupt entry, containing stored interrupt
		;	 NZ if hard case: ds:si-> our IISP interrupt entry, es:di-> IISP interrupt entry to modify
		;	CY if error (not first handler and no IISP chain until this handler)
		; CHG:	ah, es
unhookinterruptsim:
	push bx
	mov ah, 35h			; get current vector
	int 21h				; es:bx-> current interrupt handler
	cmp si, bx			; our pointer ?
	jne .hard

	push ax
	push si
	mov si, ds
	mov ax, es
	 cmp si, ax			; our segment ?
	pop si
	pop ax
	jne .hard

	and ah, 00h			; NC, ZR
	pop bx
	retn
.hard:
			; ds:si-> IISP interrupt entry
			; es:bx-> current interrupt entry
			; al = interrupt number
			;
			; CY if error
			; NC, NZ if no error,
			;  ds:si-> our IISP interrupt entry, es:di-> IISP interrupt entry to modify
	cmp word [es:bx], 10EBh		; (jump to entry's actual code) jmp short $+12h ?
	jne .fail
					; [ es:bx + 2 ] = 4 byte next entry
	cmp word [es:bx +6], "KB"	; (signature) "KB"/424Bh ?
	jne .fail
	cmp byte [es:bx +9], 0EBh	; (jump to far call "hardware reset") jmp short ... ?
					; [ es:bx + 11 ] = 7 byte "reserved" (don't look at)
	jne .fail

	cmp [es:bx +2], si		; our pointer ?
	jne .next			; no, try next -->
	push si
	mov si, ds
	 cmp [es:bx +4], si		; our segment ?
	pop si
	je .found			; yes -->
.next:
	les bx, [es:bx +2]		; get next interrupt entry
	jmp short .hard

.found:				; found reference to our interrupt handler
	mov di, bx			; es:di-> IISP interrupt entry that references our's
	or ah, 0FFh			; NC, NZ
	pop bx
	retn

.fail:				; IISP incompatible TSR between current interrupt entry and our entry
	stc
	pop bx
	retn

install:		; ds = handler segment inside cs
			; bp = option selector
	d4bp
	mov ax, cs
	add ax, 10h+transient_size_p
	mov ds, ax		; ds = handler segment (could have changed if /n switch and already resident)

%if _RXANSI_I2F
	mov cx, rxansi_handler_i2f_size_p
%else
	mov cx, rxansi_handler_noi2f_size_p
%endif

%if _RXANSI_OPTI2F
	test bp, _installi2f	; install i2f ?
	jz .noi2f
	add cx, byte rxansi_handler_i2f_size_p - rxansi_handler_noi2f_size_p

			; exchange 3 byte, from rxansi_amisintr.i2d to .i2f;
			; .i2f follows .i2d (.i2d +3 = .i2f).
	push word [rxansi_amisintr.i2d]		; first two bytes of .i2d
	push word [rxansi_amisintr.i2d+2]	; last byte of .i2d, first byte of .i2f
	push word [rxansi_amisintr.i2f+1]	; last two bytes of .i2f

	pop word [rxansi_amisintr.i2d+1]	; to last two bytes of .i2d
	pop ax
	mov [rxansi_amisintr.i2d], ah		; to first byte of .i2d
	mov [rxansi_amisintr.i2f+2], al		; to last byte of .i2f
	pop word [rxansi_amisintr.i2f]		; to first two bytes of .i2f
.noi2f:
%endif

	mov ah, 0Fh
	int 10h			; get video mode etc, bh = active video page
	mov ah, 08h		; read character and attribute at cursor position
	int 10h
	mov byte [rxansi_attrib], ah	; inherit current attribute

	mov bx, rxansi_ctrl0	; ds:bx -> ctrl0
	call changeflags	; set flags from bp


	test bp, _holdenvironment ; don't free environment ?
	jnz .dontfreeenvironment ; yes -->

	mov ah, 49h
	mov es, word [ cs:2Ch ]
	int 21h			; Free our environment
.dontfreeenvironment:

	mov sp, ..@stack	; change to stack inside now unused code

	mov ah, 4Ah
	push cs
	pop es			; Resize the segment allocated to this program
	mov bx, 10h+transient_size_p+rxansi_handler_size_p	; 10h is the size of the PSP
	int 21h

			; The DOS memory allocation strategy classes are actually "Low then high", "High only"
			; and "High then low". RBIL tells trash regarding that. We've to disable the UMB link
			; to insure only low memory is allocated. Normally, the UMB link is already disabled
			; when the program starts. But some programs, like CMDEDIT 3.21, are buggy and leave
			; the UMB link enabled after they used UMBs.
	mov bx, 1		; enable UMB link
	mov dx, 0081h		; best fit, high then low strategy
	test bp, _installlow	; use settings for high ?
	jz .high		; indeed -->
	mov bl, 0		; disable UMB link (bh already 0)
	mov dl, 01h		; best fit, low strategy (dh already 0)

.high:
	mov ax, 5802h
	int 21h
	xor ah, ah		; some "DOS" only return al
	push ax			; save UMB link
	mov ax, 5803h
	int 21h			; set UMB link (disable for low, enable for high)

	mov ax, 5800h
	int 21h
	push ax			; save strategy
	mov ax, 5801h
	mov bx, dx
	int 21h			; set strategy (low for low, high then low for high)
	jnc .highsupported
			; If the current "DOS" is < 5 (or: doesn't support UMBs),
			; the low allocation strategy is used instead:

	mov ax, 5801h		; filled with error code, reset to function number
	mov bl, 01h		; set to best fit, low strategy
	int 21h			; set strategy

.highsupported:
	mov ah, 48h
	mov bx, cx		; size we want (paragraphs)
	int 21h
	jnc .memoryallocated
	mov dx,msg.nomemory
.abort_msg:
	jmp abort_msg

		; ax = allocated memory block
		; cx = size of required memory block in paragraphs
		; ds = segment of resident (behind transient)
.memoryallocated:
	mov es, ax
		; es = ax
	test bp, _forcetoprelocation
	jnz .relocate		; don't test if behind program -->
	mov bx, cs
	add bx, 10h+transient_size_p+rxansi_handler_size_p +1
	cmp ax, bx		; did we allocate the block just behind the PSP block ?
	je .relocate		; yes, relocate -->
.dontrelocate_1:		; else no need to relocate
	jmp .dontrelocate

.relocate:

%if 0

If the allocated block is located behind the PSP block, a gap of the PSP size, transient
size and resident size (10h+transient_size_p+rxansi_handler_size_p) would be created by
using that block. Thus the allocated block is freed, the transient code is relocated into
a newly allocated block and the resident code is relocated to where the transient code
was before (behind the PSP). The PSP and the relocated transient code are both freed by
the termination but the newly created MCB of the relocated resident code (at cs+10h)
isn't. This still leaves a gap, but it's just 256 byte sized (PSP size).

This relocation helps to decrease the "gap size" if the program is loaded in the LMA and
then installs the resident code there as well (as usual if there's no UMA). The PSP
itself isn't relocated. (Relocating the PSP wouldn't create any gap at all, but it may
not work on all systems. And you've to change all allocated MCBs and the current DOS PSP
manually.) Those "good old" TSRs don't create any gap. But it's better to have a 256 byte
gap than to hold a useless PSP resident in the same gap.

If, however, the free memory above the allocated memory is _smaller_ than the gap created
by relocating the resident part behind the PSP, it's instead relocated to the end of that
memory so that no gap stays in memory.

Some examples:

Seg.  Usage
0000: 64 KiB used by DOS and other programs
1000: PSP
1010: Transient code
1090: Resident code
10C7: Free memory	<-- Memory allocated first is in this block
The transient code is relocated to somewhere else. The resident code is copied down from
segment 1090h to 1011h (at 1010h, there's the newly created MCB). The memory behind the
resident code is freed; so (with installing Int2F) memory from 1039h is free. This
results in the following memory map after the transient code terminated:
0000: 64 KiB used by DOS and other programs
1000: Gap (size of the PSP)
1010: Resident code
1048: Free memory

Seg.  Usage
0000: 64 KiB used by DOS and other programs
1000: PSP
1010: Transient code
1090: Resident code
10C7: Free memory	<-- Memory allocated first is in this block
1100: Used by another program
1300: Free memory (up to 9FFF)
The resident code is relocated to 10C8h instead, because the gap between the resident
code and the other program above is then filled. This results in the following memory
map:
0000: 64 KiB used by DOS and other programs
1000: Gap
10C8: Resident code
1100: Used by another program
1300: Free memory (up to 9FFF)

%endif

	d4bp
	push ds
	push cx
	mov ah, 4Ah		; resize memory
	mov bx, 0FFFFh		; will not succeed
	int 21h			; but returns largest possible in bx
			; Note:	DOS can actually change the size to the largest possible!
	jnc .abort_msg		; succeeded?!
	sub bx, byte 10h
	test bp, _forcetoprelocation
	jnz .relocate_beforenext
	cmp cx, bx
	jbe .relocate_behindpsp

.relocate_beforenext:
	d4 push dx
	d4 mov dx, msg.relocbeforenext
	d4 call disp_msg
	d4 pop dx
	d4bp
	add bx, byte 10h
	mov ah, 4Ah
	int 21h			; now really resize to largest possible
	sub bx, cx
	mov ax, es
	jmp short .relocate_common

.relocate_behindpsp:
	d4 push dx
	d4 mov dx, msg.relocbehindpsp
	d4 call disp_msg
	d4 pop dx
	d4bp
	mov ah, 49h		; free memory in es (allocated block)
	int 21h
	mov ah, 48h		; allocate memory for relocation of transient
	mov bx, transient_size_p
	int 21h
.abort_msg_c:
	jc .abort_msg
	mov es, ax
	xor di, di
	mov si, transient	; = 100h
	mov cx, transient_size_w
	repz cs movsw		; copy transient from cs:si-> to es:di-> (including unused code as stack)
	sub ax, byte 10h	; adjust because transient expects a PSP in front of it
	push ax			; new ss
	push ax			; new cs
	mov ax, .relocated
	push ax			; new ip
	mov ax, cs		; pass to work on it after relocation
	retf

.relocated:
	pop ss			; change stack to relocated transient (sp is still valid!)
	mov bx, 10h +1

		; ax = allocated block
		; bx = paragraph of allocated block used
		;
		; ax: PSP/lower part of allocated block
		; ax+bx: free memory (former transient)/upper part of allocated block
		;
		; This code splits the memory block at ax in two blocks. The lower block
		; (the PSP/unused) is freed by the termination of the transient code.
.relocate_common:
	dec ax
	mov ds, ax		; former transient's MCB
	add ax, bx
	mov es, ax		; just behind the PSP
	xor di, di
	xor si, si
	mov cx, 8
	repz movsw		; copy MCB at ds:si-> to es:di->
	sub word [es:3], bx	; adjust size to next
	jc .abort_msg_c
	mov word [es:5], cx
	mov byte [es:7], cl	; force the unused bytes to zero
	cmp byte [0], 'M'	; 'M' ?
	je .validmcb		; all right -->
	cmp byte [0], 'Z'	; not 'Z' ?
	jne .abort_msg		; error -->

.validmcb:
	dec bx
	mov word [3], bx	; it doesn't hurt if it's still 'Z' here
	mov byte [0], 'M'	; the created MCB always follows this one

	inc ax
	mov es, ax		; memory block of created MCB
	pop cx
	pop ds
	mov bx, cx
	mov ah, 4Ah
	int 21h
	jc .abort_msg_c
		; ax = created memory block (behind PSP)
		; cx = size of required memory block in paragraphs
		; es = segment of memory block that contains resident

.dontrelocate:

	pop bx
	mov ax, 5801h
	int 21h			; restore strategy

	pop bx
	mov ax, 5803h
	int 21h			; restore UMB link

			; es = Segment of allocated block
			; cx = size in paragraphs
	shl cx,1		; *2
	shl cx,1		; *4
	shl cx,1		; *8, convert paras to words
	xor di,di		; es:di-> start of allocated block
	xor si,si		; ds:si-> source inside this block
	push cx			; restore that later (repz movsw and Int2D.mm00 destroy cx)
	repz movsw

	mov ah, 00h		; start with multiplex number 00h
.loopamis:
	mov al, 00h		; installation check
	int 2Dh
	cmp al, 00h		; still 00h if free
	jz .foundamis
	inc ah			; search from bottom to top
	jnz .loopamis		; loop if zero not yet reached -->
	mov dx, msg.noamisnumber; abort if no free AMIS multiplex number left
	jmp abort_msg

.foundamis:
			; ah = free AMIS multiplex number
	pop cx			; restore word counter
	mov byte [ es:rxansi_amisnum ], ah	; Set multiplex number

			; All error conditions must be checked above because we'll now
			; modify the MCB so that it won't be deallocated when terminating.
	mov ax, es
	dec ax
	mov ds, ax		; ds = MCB of allocated block
	inc ax			; ax = allocated block!
	mov word [0008h], "RX"
	mov word [000Ah], "AN"
	mov word [000Ch], "SI"
	and word [000Eh],byte 0	; Force MCB string to "RXANSI",0
	mov word [0001h], ax	; Set owner to itself

			; ax = segment of resident copy
			; cx = size of copied handler in words
	cmp ax, 0A000h		; is that high ?
	mov dx, msg.highloadedusing
	jae .highloaded		; of course -->
	mov dx, msg.loadedusing	; no, other message

.highloaded:
	call disp_msg		; "[high ]loaded using "
	xchg ax, cx		; cx = allocated block, ax = word size
	shl ax, 1		; *2, convert words to bytes
	add ax, byte 16		; also count the MCB (most MEMs do that too)
	call disp_ax_dec	; "<num>"
	mov dx, msg.byte
	call disp_msg		; " bytes."

	mov ds, cx		; ds = allocated block
	mov bx, rxansi_amisintr	; ds:bx -> AMIS interrupt list
	mov dx, hookinterrupt
	mov cx, ..@exit
	jmp loopamisintr	; hook all interrupts in list, return to ..@exit

		; INP:	ds:si-> IISP interrupt entry
		;	al = interrupt number
		; OUT:	NC
		; CHG:	ah, es
hookinterrupt:
	push bx
	push dx
	mov ah, 35h
	int 21h			; get old handler
	mov [si +2], bx
	mov [si +4], es		; store

	mov ah, 25h
	mov dx, si		; ds:dx-> interrupt entry
	int 21h			; set new handler
	pop dx
	pop bx
	clc
	retn

onoff:
	mov dx, msg.off
	jz .off
	mov dx, msg.on
.off:				; fall through and display either
disp_msg:
	 push ds
	 push ax
	push cs
	pop ds
	mov ah, 09h
	int 21h
	 pop ax
	 pop ds
	retn


disp_al_hex:
		push cx
		mov cl, 4
		ror al, cl
		call .nibble
		ror al, cl
		pop cx
.nibble:
		push ax
		and al, 0Fh
		add al, '0'
		cmp al, '9'
		jbe .isdigit
		add al, 'A'-('9'+1)
.isdigit:
		xchg dl, al
		mov ah, 02h
		int 21h
		pop ax
		retn

	; Following call: Display number in ax decimal
	; all registers preserved except dx
disp_ax_dec:			; ax (no leading zeros)
	; In: number in ax
	; Out: displayed
		push bx
		xor bx, bx
.pushax:
		push ax
.pushend:
		or bl, bl
		jz .nobl
		sub bl, 5
		neg bl
.nobl:
		push cx
		mov cx, 10000
		call .divide_out
		mov cx, 1000
		call .divide_out
		mov cx, 100
		call .divide_out
		mov cl, 10
		call .divide_out
							; (Divisor 1 is useless)
		add al, '0'
		xchg dl, al
		mov ah, 02h
		int 21h
		pop cx
		pop ax
		pop bx					; Caller's register
		retn

.divide_out:
	; In: ax = number
	;     cx = divisor
	; Out: ax = remainder of operation
	;      result displayed
		push dx
		xor dx, dx
		div cx				; 0:ax / cx
		push dx				; remainder
		dec bl
		jnz .nobl2
		or bh, 1
.nobl2:
		or bh, al
		jz .leadingzero
		add al, '0'
		xchg dl, al
		mov ah, 02h
		int 21h				; display result
 .leadingzero:
		pop ax				; remainder
		pop dx
		retn

msg:
			; The version string is international
.name:		db "RxANSI ",RXANSI_VERP_STR,"beta1: ",36
%if _RXANSI_DEU
			; German messages
			;
			; The german messages are currently up-to-date.
			; Remove this comment if you change the english
			; messages without updating the german messages.
.highloadedusing:db "hoch"
.loadedusing:	db      "geladen, Speicherbedarf ",36
.byte:		db " Byte.",13,10,36
.notinstalled:	db "nicht installiert.",13,10,36
.cantuninstall:	db "Entfernen fehlgeschlagen",36
.interrupts:	db ", Interrupt(s) ",36
.comma:		db ", ",36
.hooked:	db " wurde(n) AMIS-inkompatibel umgeleitet.",13,10,36
.incompatible:	db "inkompatible Version installiert.",13,10,36
.already:	db "bereits installiert.",13,10,36
.removed:	db "entfernt.",13,10,36
.nomemory:	db "zu wenig Speicher.",13,10,36
.noamisnumber:	db "keine freie AMIS-Multiplex-Nummer.",13,10,36
.stateansi:	db "Status ANSI ",36
.stateless:	db ", LESS ",36
.on:		db "EIN (+)",36
.off:		db "AUS (-)",36
.help:		db "Kombinierter ANSI- und LESS-Treiber. AMIS-v3.6-konform.",13,10
		db "Public Domain von Henrik Haftmann und Christian Masloch",13,10
		db "Optionen: (kein) Installiere oder zeige Status",13,10
		db "        /c       Installiere in konventionellen Speicher",13,10
%if _RXANSI_OPTI2F
		db "        /s       Installiere sicher (leite auch Interrupt 2Fh um)",13,10
%endif
		db "        /n       Installiere neu selbst wenn bereits installiert",13,10
		db "        /a{+|-}  ANSI-Treiber ein/aus, Standard EIN",13,10
		db "        /l{+|-}  LESS-Treiber ein/aus, Standard EIN",13,10
		db "        /u       Entferne vom Speicher",13,10
		db "        /?, /h   Diese Hilfe",13,10
		db 13,10
		db "Die Optionen /c"
%if _RXANSI_OPTI2F
		db		  ", /s"
%endif
		db		      ", /n, /a und /l k",148,"nnen kombiniert werden.",13,10
		db "Das Optionszeichen - wird ebenfalls erkannt."
		; character 148 is the German 'O' umlaut in CP437/CP850
%else
			; English messages
.highloadedusing:db "high "
.loadedusing:	db       "loaded using ",36
.byte:		db " byte.",13,10,36
.notinstalled:	db "not installed.",13,10,36
.cantuninstall:	db "can't uninstall",36
.interrupts:	db ", Interrupt(s) ",36
.comma:		db ", ",36
.hooked:	db " hooked AMIS incompatible.",13,10,36
.incompatible:	db "incompatible version installed.",13,10,36
.already:	db "already installed.",13,10,36
.removed:	db "removed.",13,10,36
.nomemory:	db "out of memory.",13,10,36
.noamisnumber:	db "no free AMIS multiplex number.",13,10,36
.stateansi:	db "State ANSI ",36
.stateless:	db ", LESS ",36
.on:		db "ON (+)",36
.off:		db "OFF (-)",36
.help:		db "Combined ANSI and LESS driver. AMIS v3.6 compliant.",13,10
		db "Public Domain by Henrik Haftmann and Christian Masloch",13,10
		db "Options: (no)    Install or show state",13,10
		db "        /c       Install into conventional memory",13,10
%if _RXANSI_OPTI2F
		db "        /s       Install save (hook interrupt 2Fh as well)",13,10
%endif
		db "        /n       Install new even if already installed",13,10
		db "        /a{+|-}  ANSI functions on/off, default ON",13,10
		db "        /l{+|-}  LESS functions on/off, default ON",13,10
		db "        /u       Remove from memory",13,10
		db "        /?, /h   This help message",13,10
		db 13,10
		db "The options /c"
%if _RXANSI_OPTI2F
		db		 ", /s"
%endif
		db		     ", /n, /a and /l can be combined.",13,10
		db "The switch character - is also accepted."
%endif
d4 .relocbeforenext:	db "Relocating resident to before next",13,10,36
d4 .relocbehindpsp:	db "Relocating resident to behind PSP",13,10,36
.crlf:		db 13,10,36

parameters:			; Valid switches on the command line, followed
				; by an option selector byte with one bit set.
%if _RXANSI_OPTI2F
	db "S", _installi2f
%endif
	db "C", _installlow
	db "N", _installnew
	db "U", _uninstall
	db "A", _ansistate
	db "L", _lessstate
			; These undocumented options change relocation behaviour:
	db "E", _holdenvironment
	db "T", _forcetoprelocation
	db 0FFh			; table end marker

			; The help switches are not in this table (and neither in the option
			; selector) because on invalid switches, we'll show the help anyway.

	endarea transient


	section RESIDENT vstart=0 align=16
				; The handler is later copied into a seperate memory block.
				; The vstart= value tells NASM to use all addresses as if this
				; location is actually at vstart. (In this case, like if it's
				; at the start of the segment, not at 100h+start_size.)


%include "RXANSI.ASM"

%if _UPXPAD && (transient_size_p*16 + rxansi_handler_size) < 3062
		times 3062 -(transient_size_p*16 + rxansi_handler_size) db 90h
%endif